import { namehash } from '@ethersproject/hash';
import { Contract } from '@ethersproject/contracts';
import { getDefaultProvider } from '@ethersproject/providers';
import assert from 'assert';

const ABI = [
    {
        anonymous: false,
        inputs: [
            {
                indexed: true,
                internalType: 'bytes32',
                name: 'node',
                type: 'bytes32'
            },
            {
                indexed: true,
                internalType: 'string',
                name: 'indexedKey',
                type: 'string'
            },
            {
                indexed: false,
                internalType: 'string',
                name: 'key',
                type: 'string'
            }
        ],
        name: 'TextChanged',
        type: 'event'
    },
    {
        inputs: [
            {
                internalType: 'bytes32',
                name: 'node',
                type: 'bytes32'
            },
            {
                internalType: 'string',
                name: 'key',
                type: 'string'
            },
            {
                internalType: 'string',
                name: 'value',
                type: 'string'
            }
        ],
        name: 'setText',
        outputs: [],
        stateMutability: 'nonpayable',
        type: 'function'
    },
    {
        inputs: [
            {
                internalType: 'bytes4',
                name: 'interfaceID',
                type: 'bytes4'
            }
        ],
        name: 'supportsInterface',
        outputs: [
            {
                internalType: 'bool',
                name: '',
                type: 'bool'
            }
        ],
        stateMutability: 'pure',
        type: 'function'
    },
    {
        inputs: [
            {
                internalType: 'bytes32',
                name: 'node',
                type: 'bytes32'
            },
            {
                internalType: 'string',
                name: 'key',
                type: 'string'
            }
        ],
        name: 'text',
        outputs: [
            {
                internalType: 'string',
                name: '',
                type: 'string'
            }
        ],
        stateMutability: 'view',
        type: 'function'
    },
    {
        inputs: [
            {
                internalType: 'bytes32',
                name: 'node',
                type: 'bytes32'
            },
            {
                internalType: 'uint256',
                name: 'coinType',
                type: 'uint256'
            }
        ],
        name: 'addr',
        outputs: [
            {
                internalType: 'bytes',
                name: '',
                type: 'bytes'
            }
        ],
        stateMutability: 'view',
        type: 'function'
    }
];

const ENDPOINT = 'https://api.thegraph.com/subgraphs/name/ensdomains/ens';
const QUERY = `
query($domain: String!) {
  domains(where:{name: $domain}) { 
    resolvedAddress {
      id
    }
    resolver {
      texts
      coinTypes
    }
    owner {
      id
    }
  }
}
`;
const request = async (endpoint, query, variables, fetchOptions = {}) => {
    const res = await fetch(endpoint, {
        ...fetchOptions,
        body: JSON.stringify({ query, variables }),
        method: 'POST'
    });
    assert.strictEqual(200, res.status);
    const json = await res.json();
    return json.data.domains[0];
};
const ADDRESS_REGEX = /^0x[a-fA-F0-9]{40}$/;
/**
 *
 * @param provider Ethereum provider
 * @param contractAddress ENS resolver contract address
 * @returns
 */
const getENS = (opts) => {
    const provider = opts?.provider || getDefaultProvider();
    const contractAddress = opts?.contractAddress || '0x4976fb03C32e5B8cfe2b6cCB31c09Ba78EBaBa41';
    const endpointUrl = opts?.endpointUrl || ENDPOINT;
    const contract = new Contract(contractAddress, ABI, provider);
    const getRecord = async (node, record) => await contract.functions.text(node, record);
    const getCoin = async (node, coinType) => await contract.functions.addr(node, coinType);
    return async function getENS(_domain, fetchOptions) {
        if (!(ADDRESS_REGEX.test(_domain) || _domain?.endsWith('.eth')))
            throw new Error(`Invalid ENS domain or ethereum address: ${_domain}`);
        const domain = ADDRESS_REGEX.test(_domain) ? await provider.lookupAddress(_domain) : _domain;
        if (domain == null) {
            return {
                owner: _domain,
                address: _domain,
                domain: null,
                records: {},
                coins: {}
            };
        }
        const node = namehash(domain);
        const ens = await request(endpointUrl, QUERY, {
            domain
        }, fetchOptions);
        if (ens) {
            const { resolvedAddress: address, resolver, owner } = ens;
            let data = {
                owner: null,
                address: null,
                domain,
                records: {},
                coins: {}
            };
            if (owner?.id)
                data.owner = owner.id;
            if (address?.id)
                data.address = address.id;
            if (resolver?.texts) {
                for (const record of resolver.texts) {
                    const value = (await getRecord(node, record))[0];
                    if (record.startsWith('com.') || record.startsWith('vnd.') || record.startsWith('org.')) {
                        data.records[record.slice(record.indexOf('.') + 1)] = value;
                    }
                    else {
                        data.records[record] = value;
                    }
                }
            }
            if (resolver?.coinTypes) {
                for (const coinType of resolver.coinTypes) {
                    data.coins[coinType] = (await getCoin(node, coinType))[0];
                }
            }
            return data;
        }
        else {
            return {
                owner: null,
                address: null,
                records: {},
                domain: null,
                coins: {}
            };
        }
    };
};

export { getENS };
